# class surfit is a wrapper class for dierchk's surfit and bispev routines.
from Math import zeros, ones, sqrt, array
from _surfit import *
SURFIT_ERROR = 'SURFIT error'

class surfit:
    def __init__(self, x, y, z, xb=None, xe=None, yb=None, ye=None,
                 w=None, kx=3, ky=3, nxest=None, nyest=None):
        # data, the validity is left for the underlying surfit to check 
        self.x = array(x).astype('f').flat
        self.y = array(y).astype('f').flat
        self.z = array(z).astype('f').flat
        self.xb = xb and xb or min(self.x)
        self.xe = xe and xe or max(self.x)
        self.yb = yb and yb or min(self.y)
        self.ye = ye and ye or max(self.y)
        self.w = w and w or ones(self.m, 'f')
        self.w = array(self.w).astype('f')
        self.m = self.x.shape[0]
        if self.m != self.y.shape[0] or self.m != self.z.shape[0] \
	   or self.m != self.w.shape[0]:
            raise SURFIT_ERROR, 'Incompatible input data demension'

        # the config of the bivar spline
        self.kx = kx
        self.ky = ky
        self.nxest = nxest and nxest or self.kx + 1 + int(sqrt(self.m/2))
        self.nyest = nyest and nyest or self.ky + 1 + int(sqrt(self.m/2))
	self.nx = 0
	self.ny = 0
	
        # knots and coefficients
        self.nmax = max(self.nxest, self.nyest)
        self.tx = zeros(self.nmax, 'f')
        self.ty = zeros(self.nmax, 'f')
        u = self.nxest - self.kx - 1
        v = self.nyest - self.ky - 1
        self.c = zeros(u*v, 'f')
        
        # working space
        km = max(self.kx, self.ky) + 1
        ne = max(self.nxest, self.nyest)
        bx = self.kx*v + self.ky + 1
        by = self.ky*u + self.kx + 1
        if bx <= by:
            b1 = bx; b2 = b1 + v - self.ky
        else:
            b1 = by; b2 = b1 + u - self.kx
        self.lwrk1 = u*v*(2 + b1 + b2) + 2*(u + v + km*(self.m + ne) + ne - self.kx - self.ky) + b2 + 1     + 1000
        self.wrk1 = zeros(self.lwrk1, 'f')
        self.lwrk2 = u*v*(b2 + 1) + b2
        self.wrk2 = zeros(self.lwrk2, 'f')
        self.kwrk = self.m + (self.nxest - 2*self.kx - 1)*(self.nyest - 2*self.ky - 1)
        self.iwrk = zeros(self.kwrk, 'i')

        # misc
        self.eps = 0.1e-5		# single precision
            

    def fit(self, iopt, s):
        self.nx, self.ny, fp, ierr = surfit_(iopt, self.m, self.x, self.y, self.z, self.w, self.xb, self.xe, self.yb, self.ye, self.kx, self.ky, s, self.nxest, self.nyest, self.nmax, self.eps, self.nx, self.tx, self.ny, self.ty, self.c, self.wrk1, self.lwrk1, self.wrk2, self.lwrk2, self.iwrk, self.kwrk) 

	if ierr == 0:
            print 'Fit success. abs(fp-s)/s=', abs(fp-s)/s
        elif ierr == -1:
            print 'Interpolate success.'
        elif ierr == -2:
            print 'Weighted least square fit success. Upbound for s is found:', fp
        elif ierr < -2:
            # depends on eps very much
            print 'Warning: rank deficient system with rank:', -ierr, 'and rank deficuency:', (self.nx - self.kx - 1)*(self.ny - self.ky - 1) + ierr
        else:
            raise SURFIT_ERROR, ierr

	return [ierr, fp]


    def eval(self, x, y):
        x = array(x).astype('f').flat
        y = array(y).astype('f').flat
        mx = x.shape[0]
        my = y.shape[0]
        z = zeros((mx, my), 'f')

        # working space
        lwrk = mx*(self.kx + 1) + my*(self.ky + 1)
        wrk = zeros(lwrk, 'f')
        kwrk = mx + my
        iwrk = zeros(kwrk, 'i')
        
        ierr = bispev_(self.tx, self.nx, self.ty, self.ny, self.c, self.kx, self.ky, x, mx, y, my, z, wrk, lwrk, iwrk, kwrk)

        if ierr == 10:
            raise SURFIT_ERROR, 'Incompatible input data'

	return z
    

    def knot(self):
	return [self.tx[:self.nx], self.ty[:self.ny]]


    def coeff(self):
	ns = (self.nx - self.kx - 1)*(self.ny - self.ky - 1)
	return self.c[:ns]
    

def test():
    import TableIO
    [x, y, z] = TableIO.readColumns('dasurf', '')
    w = ones(80)/0.102569e-01
    ss = surfit(x, y, z, -2, 2, -2, 2, w)
    
    ss.fit(0, 900000.)
    print 'Knots:\n', ss.knot()
    print 'Coeff:\n', ss.coeff()
    
    for s in [200, 80, 20]:
	print '\ns=', s
	try:
	    ss.fit(1, s)
	    print 'Knots:\n', ss.knot()
	    print 'Coeff:\n', ss.coeff()
	except: pass
	
    ss.fit(-1, 80)
    print '\ns=', 80
    print 'Knots:\n', ss.knot()
    print 'Coeff:\n', ss.coeff()


if __name__ == '__main__':
    test()
